import java.util.Vector;
import java.util.Collections; 
import java.awt.Toolkit;

/******************************************************************************
*  Arena of battling gladiators (fighters).
*  Meant to assess average abilities & hp at advanced levels. 
*
*  @author   Daniel R. Collins (dcollins@superdan.net)
*  @since    2014-05-20
*  @version  1.05
******************************************************************************/

public class Arena {

	//--------------------------------------------------------------------------
	//  Fields
	//--------------------------------------------------------------------------
	enum ArenaType {ManVsMan, ManVsMonster};
	final ArenaType ARENA_TYPE = ArenaType.ManVsMonster;
	final int NUM_FIGHTERS = 100000;
	final int NUM_CYCLES = 200;
	final int START_LEVEL = 0;
	final int HIGH_LEVEL = 5;
	final int MAX_LEVEL_STATS = 100;
	final int MAX_ROUNDS = 100;
	final boolean HIGH_LEVEL_REPORTS = false;
	final static boolean PRINT_ALL = false;
	final boolean SORT_LIST = false;
	Vector<Persona> fighterList;

	//--------------------------------------------------------------------------
	//  Constructors
	//--------------------------------------------------------------------------

	/**
	*  Constructor.
	*/
	public Arena () {
		fighterList = new Vector<Persona>();
		recruitNewFighters();
	}

	//--------------------------------------------------------------------------
	//  Methods
	//--------------------------------------------------------------------------

	/**
	*  Run the arena's top-level algorithm.
	*/
	void runArena () {
		for (int cycle = 0; cycle < NUM_CYCLES; cycle++) {
			organizeList();
			fightDuels();
			bringOutYourDead();
			setAllToMaxHitPoints();
			recruitNewFighters();
		}
	}

	/**
	*  Organizes the list by either sorting or shuffling.
	*/
	void organizeList () {
		if (SORT_LIST)
			Collections.sort(fighterList, Collections.reverseOrder());
		else
			Collections.shuffle(fighterList);
	}

	/**
	*  Fight duels for all fighters in list.
	*/
	void fightDuels () {
		if (ARENA_TYPE == ArenaType.ManVsMan)
			fightDuelsManVsMan();
		else
			fightDuelsManVsMonster();
	}

	/**
	*  Duel each pair of fighters.
	*/
	void fightDuelsManVsMan () {
		for (int i = 0; i < fighterList.size()-1; i+=2) {
			fightDuel(fighterList.get(i), fighterList.get(i+1));
		}	
	}

	/**
	*  Remove all dead from the fighter list.
	*/
	void bringOutYourDead () {
		int idx = 0;
		while (idx < fighterList.size()) {
			if (fighterList.get(idx).isDead())
				fighterList.remove(idx);
			else
				idx++;					
		}
	}

	/**
	*  Set everyone to max hit points.
	*/
	void setAllToMaxHitPoints () {
		for (Persona p: fighterList) {
			p.setHitPointsToMax();
		}	
	}

	/**
	*  Fill out the fighter list.
	*/
	void recruitNewFighters() {
		ClassInfo info = ClassList.getInfo("Fighter");
		while (fighterList.size() < NUM_FIGHTERS) {
			fighterList.add(new Persona(null, "Human", 
				Alignment.Neutral, info, START_LEVEL));
		}
	}

	/**
	*  Fight one paired duel.
	*/
	void fightDuel (Creature p1, Creature p2) {

		// Flip coin for initiative
		if (Math.random() > 0.5) {
			Creature swap = p1;
			p1 = p2;
			p2 = swap;
		}	

		// Strike blows
		while (true) {
			strike(p1, p2);
			if (p2.isDead()) {
				awardXP(p1, p2);
				break;
			}
			strike(p2, p1);
			if (p1.isDead()) {
				awardXP(p2, p1);
				break;			
			}
		}
	}

	/**
	*  Have one creature strike another in melee.
	*/
	public void strike(Creature attacker, Creature defender) {
		Dice d20 = new Dice(20);
		Attack attack = attacker.getAttack();
		for (int i = 0; i < attack.getRate(); i++) {
			int naturalRoll = d20.roll();
			int totalRoll = naturalRoll + attack.getBonus() + defender.getAC();
			if (naturalRoll == 20 || totalRoll >= 20) {
				int damage = attack.getDamage().roll();
				defender.takeDamage(damage);
			}
		}
	}

	/**
	*  Award XP to one creature for defeating another.
	*  Note: No reduction for difference in levels.
	*/
	void awardXP (Creature winner, Creature loser) {
		int HD = loser.getHitDiceNum();
		int xpAward = totalXPAward(HD);
		winner.addXP(xpAward);
	}

	/**
	*  Compute total XP award (monster + expected treasure).
	*  Treasure accounts for about 95% of XP at 1st level, 
	*  -1% per added level; here we invert that.
	*  See my DeltasDnD blog 2014-02-25.
	*/
	int totalXPAward (int HD) {
		double treasPercent = 0.95 - 0.01 * (HD-1);	
		int totalXP = (int) 
			Math.round(1/(1-treasPercent) * baseXPAward(HD));
		return totalXP;
	}
	
	/**
	*  Compute the base XP award by HD (as per Sup-I).
	*  No partial-HD or special abilities are included here.
	*/
	int baseXPAward (int HD) {
		if (HD < 0)
			return 0;
			
		switch (HD) {
			case 0: return 5;
			case 1: return 10;
			case 2: return 20;
			case 3: return 35;
			case 4: return 75;
			case 5: return 175;
			case 6: return 275;
			case 7: return 450;
			case 8: return 650;
			case 9: case 10: return 900;
			case 11: case 12: return 1100;
			case 13: case 14: return 1350;
			case 15: case 16: return 1650;
			case 17: case 18: case 19: case 20: return 2000;
			default: return 2500; 
		}
	}

	/**
	*  Print all of the fighters (for testing small groups).
	*/
	void printAllFighters() {
		if (PRINT_ALL) {
			for (Persona p: fighterList) {
				System.out.println(p);
			}
		}
	}

	/**
	*  Generate & print statistics for all the fighters.
	*/
	public void printStatistics () {

		// Create the stat bins
		StatBin bin[] = new StatBin[MAX_LEVEL_STATS];
		for (int i = 0; i < MAX_LEVEL_STATS; i++)
			bin[i] = new StatBin();

		// Add all fighters data into stat bins
		for (Persona p: fighterList) {
			int level = p.getLevel();
			if (0 <= level && level < MAX_LEVEL_STATS) {
				bin[level].addPersona(p);
			}
		}

		// Print out the bin statistics
		for (int i = 0; i < MAX_LEVEL_STATS; i++) {
			if (bin[i].getNum() > 0) {
				System.out.print((i) + "\t");
				bin[i].printAllStats();
			}
		}
		System.out.println();
		
		// Print out selected data
//		bin[3].printAllData();
	}

	/**
	*  Duel each fighter against a random monster.
	*/
	void fightDuelsManVsMonster () {
		MonsterRandomizer ml = MonsterRandomizer.getInstance();
		for (Persona fighter: fighterList) {
			int fighterLevel = fighter.getLevel();
			Creature monster = ml.randomMonsterOnLevel(fighterLevel); 
			fightDuel(fighter, monster);
			
			// Print high-level fight reports
			if (HIGH_LEVEL_REPORTS) {
		 		if (fighter.getHitDiceNum() >= HIGH_LEVEL 
 						&& monster.getHitDiceNum() >= HIGH_LEVEL) {
	 				System.out.println(fighter);
					System.out.println("  vs. " + monster);
	 			}
			}
		}
	}

	/**
	*  Main test method.
	*/
	public static void main (String[] args) {
		Arena a = new Arena();
		a.runArena();
		System.out.println();
		a.printStatistics();
		a.printAllFighters();
		Toolkit.getDefaultToolkit().beep();
	}
}

