/******************************************************************************
*  Dice group for random rolls.
*
*  @author   Daniel R. Collins (dcollins@superdan.net)
*  @since    2014-05-20
*  @version  1.1
******************************************************************************/

public class Dice {
	//--------------------------------------------------------------------------
	//  Fields
	//--------------------------------------------------------------------------

	/** Number of dice. */
	private int number;
	
	/** Sides on the dice. */
	private int sides;

	/** Multiplier to the dice (negative = divisor). */
	private int multiplier;
	
	/** Addition to the dice (negative = subtraction). */
	private int addition;

	//--------------------------------------------------------------------------
	//  Constructors
	//--------------------------------------------------------------------------

	/**	
	*  Constructor (one die only).
	*/
	Dice (int sides) {
		this(1, sides, 1, 0);
	}

	/**	
	*  Constructor (number, sides).
	*/
	Dice (int num, int sides) {
		this(num, sides, 1, 0);
	}

	/**	
	*  Constructor (number, sides, addition).
	*/
	Dice (int num, int sides, int add) {
		this(num, sides, 1, add);
	}

	/**	
	*  Constructor (full fields).
	*/
	Dice (int number, int sides, int mul, int add) {
		this.number = number;
		this.sides = sides;
		this.multiplier = mul;
		this.addition = add;	
	}

	/**	
	*  Constructor (read from string descriptor).
	*/
	Dice (String s) {
		int dIdx = s.indexOf('d');

		// Fixed value
		if (dIdx == -1) {
			number = 0;
			sides = 0;
			multiplier = 1;
			addition = Integer.parseInt(s);
		}		

		// Normal dice string
		else {
			number = dIdx > 0 ? Integer.parseInt(s.substring(0, dIdx)) : 1;
			sides = integerAfter(s, 'd', 0);
			multiplier = integerAfter(s, 'x', 1);
			if (multiplier == 1) {
				multiplier = - integerAfter(s, '/', -1);
			}
			addition = integerAfter(s, '+', 0);
			if (addition == 0) {
				addition = - integerAfter(s, '-', 0);
			}
		}
	}

	//--------------------------------------------------------------------------
	//  Methods
	//--------------------------------------------------------------------------
	public int getNumber() { return number; }
	public int getSides() { return sides; }
	public int getMultiplier() { return multiplier; }
	public int getAddition() { return addition; }

	/**	
	*  Rolls the dice.
	*  @return The dice-roll, minimum 1.
	*/
	public int roll () {
		return rollPlus(0);
	}

	/**	
	*  Rolls the dice with added modifier.
	*  @return The dice-roll, minimum 1.
	*/
	public int rollPlus (int plus) {

		// Roll the dice
		int sum = 0;
		for (int i = 0; i < number; i++) {
			sum += (int)(Math.random() * sides) + 1;
		}

		// Apply modifiers
		int total;
		if (multiplier >= 0)
			total = sum * multiplier;
		else
			total = sum / (-multiplier);
		total += addition + plus;
		return (total > 0 ? total : 1);
	}

	/**
	*  Returns a string representation of this dice group.
	*/
	public String toString() {
		if (number < 1) {
			return "" + addition;		
		}
		else {
			String s = number + "d" + sides;
			if (multiplier < 0)
				s += "/" + (-multiplier);
			else if (multiplier != 1)
				s += "x" + multiplier;
			if (addition < 0)
				s += addition;
			else if (addition > 0)
				s += "+" + addition;
			return s;			
		}
	}

	/**
	*  Find the first non-digit in a string after a given index.
	*/
	public static int nextNonDigit (String s, int beginIdx) {
		for (int idx = beginIdx; idx < s.length(); idx++) {
			if (!Character.isDigit(s.charAt(idx)))
				return idx;
		}
		return s.length();
	}

	/**
	*  Return the positive integer after a given character in a string.
	*  @return The integer, or iDefault if character is not found.
	*/
	public static int integerAfter (String s, char c, int iDefault) {
		int charIdx = s.indexOf(c);
		if (charIdx > -1) {
			int endIdx = nextNonDigit(s, charIdx + 1);
			return Integer.parseInt(s.substring(charIdx + 1, endIdx));
		}
		return iDefault;
	}

	/**
	*  Print dice and roll (for testing).
	*/
	public void reportRoll () {
		System.out.println(this + " = " + roll());
	}
	
	/**
	*  Main test function.
	*/
	public static void main (String[] args) {

		// Test numerical constructors
		new Dice(6).reportRoll();
		new Dice(2, 6).reportRoll();
		new Dice(4, 6, 1).reportRoll();
		new Dice(1, 6, -7).reportRoll();
		new Dice(3, 10, 10, 0).reportRoll();
		new Dice(2, 6, 3, -5).reportRoll();
		new Dice(1, 6, -2, 0).reportRoll();
		new Dice(0, 0, 1, 2).reportRoll();
		
		// Test string constructor
		new Dice("d6").reportRoll();
		new Dice("2d6").reportRoll();
		new Dice("4d6+1").reportRoll();
		new Dice("3d10x10").reportRoll();
		new Dice("2d6x3-5").reportRoll();
		new Dice("d6/2").reportRoll();
		new Dice("2").reportRoll();
	}
}

