import java.io.IOException; 
import java.util.ArrayList;

/******************************************************************************
*  Generates random monsters, as for wandering encounters (singleton pattern).
*
*  @author   Daniel R. Collins (dcollins@superdan.net)
*  @since    2014-07-18
*  @version  1.0
******************************************************************************/

public class MonsterRandomizer {
	//--------------------------------------------------------------------------
	//  Constants
	//--------------------------------------------------------------------------

	/** Name of file with monster information. */
	final String RANDOM_LEVELS_FILE = "RandomLevels.csv";

	/** Modifier to level determination d6 roll. */
	final int ROLL_MODIFIER = -4;

	//--------------------------------------------------------------------------
	//  Fields
	//--------------------------------------------------------------------------

	/** The singleton class instance. */
	static MonsterRandomizer instance = null;

	/** Array of random level dice points. */
	int[][] randomLevelArrays;

	/** Array of monsters of each level. */
	ArrayList<ArrayList<Creature>> monsterLevelArrays;

	//--------------------------------------------------------------------------
	//  Constructors
	//--------------------------------------------------------------------------

 	/**
 	*  Constructor (read from dedicated file).
 	*/
 	protected MonsterRandomizer () throws IOException {
		readRandomLevelsTable();
		makeMonsterLevelArrays();
	}

	//--------------------------------------------------------------------------
	//  Methods
	//--------------------------------------------------------------------------

	/**
	*  Access the singleton class instance.
	*/
	public static MonsterRandomizer getInstance() {
		if (instance == null) {
			try {
				instance = new MonsterRandomizer();
			}
			catch (IOException e) {
				System.out.println("Failed to read the RandomLevels file.");
			}
		}
		return instance;
	}

	/**
	*  Read in the random levels table.
	*/
	private void readRandomLevelsTable () throws IOException {
		InfoTable table = InfoTable.readTableFile(RANDOM_LEVELS_FILE)[0];
		int rows = table.numRows();
		int cols = table.numColumns();
		randomLevelArrays = new int[rows][];
		for (int i = 0; i < rows; i++) {
			randomLevelArrays[i] = new int[cols];
			for (int j = 0; j < cols; j++) {
				String cell = table.getCell(i, j);
				randomLevelArrays[i][j] = cell.equals("More") ?
					Integer.MAX_VALUE : table.getCellAsInt(i, j);
			}
		}
	}

	/**
	*  Compile the monster level arrays.
	*/
	private void makeMonsterLevelArrays () {

		// Initialize the arrays
		int maxLevel = randomLevelArrays[0].length + 1;
		monsterLevelArrays = new ArrayList<ArrayList<Creature>>(maxLevel-1);
		for (int i = 0; i < maxLevel-1; i++) {
			 monsterLevelArrays.add(new ArrayList<Creature>());		
		}		
		
		// Sort monster list into these bins
		MonsterList monList = MonsterList.getInstance();
		for (int i = 0; i < monList.size(); i++) {
			Creature monster = monList.get(i);
			int monLevel = monster.getMonsterLevel();
			if (1 <= monLevel && monLevel <= maxLevel) {
				monsterLevelArrays.get(monLevel-1).add(monster);			
			}		
		}
	}

	/**
	*  Roll a random monster level for a given dungeon level.
	*/
	public int randomMonsterLevel (int dungeonLevel) {

		// Roll appropriate die
		int roll = new Dice(6).roll() + ROLL_MODIFIER;

		// Get array row to look at
		int row = 0;
		while (randomLevelArrays[row][0] < dungeonLevel) {
			row++;
		}

		// Scan for first nonzero cell the roll falls in
		int[] levelArray = randomLevelArrays[row];
		for (int i = 1; i < levelArray.length; i++) {
			int val = levelArray[i];
			if (val != 0 && roll <= val) {
				return i;
			}
		}
		
		// Error condition
		return -1;
	}

	/**
	*  Roll a new random monster of a given monster level.
	*/
	public Creature randomMonsterOfLevel (int monsterLevel) {
		ArrayList<Creature> levelArray 
			= monsterLevelArrays.get(monsterLevel-1);		
		int idx = new Dice(levelArray.size()).roll()-1;
		return levelArray.get(idx).spawn();
	}

	/**
	*  Roll a new random monster on a given dungeon level.
	*/
	public Creature randomMonsterOnLevel (int dungeonLevel) {
		int monsterLevel = randomMonsterLevel(dungeonLevel);
		return randomMonsterOfLevel(monsterLevel);
	}

	/**
	*  Main test method.
	*/
	public static void main (String[] args) {
		MonsterRandomizer mr = MonsterRandomizer.getInstance();
	
		// Five random 1st-level monsters
		for (int i = 0; i < 5; i++) {
			System.out.println(mr.randomMonsterOfLevel(1));		
		}
		System.out.println();
		
		// Five random monsters on dungeon level 5
		for (int i = 0; i < 5; i++) {
			System.out.println(mr.randomMonsterOnLevel(5));
		}
		System.out.println();
	
		// Five random monsters on dungeon level 20
		for (int i = 0; i < 5; i++) {
			System.out.println(mr.randomMonsterOnLevel(20));
		}
		System.out.println();
	}
}

