import java.util.ArrayList;

/******************************************************************************
*  Persona (player or non-player character).
*
*  @author   Daniel R. Collins (dcollins@superdan.net)
*  @since    2014-05-20
*  @version  1.0
******************************************************************************/

public class Persona extends Creature {
	//--------------------------------------------------------------------------
	//  Fields
	//--------------------------------------------------------------------------

	/** The six ability scores. */
	int [] abilityScore;

	/** List of class records gained. */
	ArrayList<ClassRecord> classRecordList;

	/** Whether we apply the prime-requisite XP bonus. */
	static final boolean APPLY_BONUS_XP = false;

	//--------------------------------------------------------------------------
	//  Constructors
	//--------------------------------------------------------------------------

	/**
	*  Constructor (single class).
	*/
	public Persona (String name, String type, Alignment align, 
			ClassInfo classInfo, int level) {
		super(9, 12, 1);

		// Basics
		this.name = name;
		this.type = type;
		this.alignment = align;
		createRandomAbilityScores();
		ClassRecord cRecord = 
			new ClassRecord(classInfo, level, getAbilityBonus(Ability.Con));
		classRecordList = new ArrayList<ClassRecord>();
		classRecordList.add(cRecord);
		int atkBonus = cRecord.attackBonus() + getAbilityBonus(Ability.Str);
		setHitPointsToMax();

		// Clunky switch on class HD (TODO: improve)
		switch (classInfo.getHitDiceType()) {
			case 8: // Fighter
				moveInches = 9;
				armorClass = 2 - getAbilityBonus(Ability.Dex); // plate & shield
				attack = new Attack("Sword", 1, atkBonus, 
					new Dice(1, 8, getAbilityBonus(Ability.Str)));
				break;

			case 6: // Thief
				moveInches = 12;
				armorClass = 7 - getAbilityBonus(Ability.Dex); // leather
				attack = new Attack("Sword", 1, atkBonus, 
					new Dice(1, 8, getAbilityBonus(Ability.Str)));
				break;

			case 4: // Wizard
				moveInches = 12;
				armorClass = 9 - getAbilityBonus(Ability.Dex); // no armor
				attack = new Attack("Dagger", 1, atkBonus, 
					new Dice(1, 4, getAbilityBonus(Ability.Str)));
				break;
		}
	}

	//--------------------------------------------------------------------------
	//  Methods
	//--------------------------------------------------------------------------

	/**
	*  Creates random ability scores.
	*/
	private void createRandomAbilityScores () {
		Dice dice = new Dice(3, 6);
		abilityScore = new int[Ability.values().length];
 		for (int i = 0; i < abilityScore.length; i++) {
 			abilityScore[i] = dice.roll();
 		}
	}

	/**
	*  Gets an ability score bonus/modifier.
	*/
	public int getAbilityBonus (Ability ability) {
		int score = getAbilityScore(ability);
		return Ability.getBonus(score);
	}

	/**
	*  Gets an ability score.
	*/
	public int getAbilityScore (Ability ability) {
		return abilityScore[ability.ordinal()];
	}

	/**
	*  Gets the ability score by index.
	*/
	public int getAbilityScoreByIndex (int index) {
		return abilityScore[index];	
	}

	/**
	*  Get maximum class level.
	*/
	public int getLevel () {
		int maxLevel = 0;
		for (ClassRecord record: classRecordList) {
			if (record.getLevel() > maxLevel)
				maxLevel = record.getLevel();
		}
		return maxLevel;
	}

	/**
	*  Returns max level as hit dice number (max over all classes).
	*  TODO: Record in hitDice field at appropriate time. (?)
	*  ## This is probably the clunkiest bit here. Also want level return? ##
	*/
	public int getHitDiceNum () {
		return Math.max(1, getLevel());
	}

	/**
	*  Returns max hit points (max over all classes).
	*  TODO: Record in maxHitPoints field at appropriate time.
	*/
	public int getMaxHitPoints () {
		int maxHP = 0;
		for (ClassRecord record: classRecordList) {
			if (record.getHitPoints() > maxHP)
				maxHP = record.getHitPoints();		
		}
		return maxHP;
	}

	/**
	*  Returns base attack bonus (max over all classes).
	*/
	public int baseAttackBonus () {
		int maxBAB = 0;
		for (ClassRecord record: classRecordList) {
			if (record.attackBonus() > maxBAB)
				maxBAB = record.attackBonus();
		}
		return maxBAB;
	}

	/**
	*  Returns total XP (sum over all classes).
	*/
	public int totalXP () {
		int sum = 0;
		for (ClassRecord record: classRecordList) {
			sum += record.getXP();
		}
		return sum;
	}

	/**
	*  Add XP to this persona.
	*  TODO: Currently adds to & updates only the first class.
	*/
	public void addXP (int XP) {
		int primeScore = primeRequisite();
		if (APPLY_BONUS_XP) {
			XP = XP + XP * Ability.bonusPercentXP(primeScore)/100;
		}
		ClassRecord record = classRecordList.get(0);
		if (record.addXP(XP)) {
			record.rollNewHitPoints(getAbilityBonus(Ability.Con));
			int atkBonus = record.attackBonus() + getAbilityBonus(Ability.Str);
			attack.setBonus(atkBonus);
		}
	}

	/**
	*  Returns the effective prime requisite score.
	*  TODO: Currently assumes fighter class, generalize for given class.
	*    Possibly shove this down into ClassInfo or ClassRecord.
	*/
	private int primeRequisite () {
		// Raw prime requisite score
		return getAbilityScore(Ability.Str);

// 		// Vol-1 modifiers to prime requisite
// 		return getScore(Ability.Str) 
// 			+ getScore(Ability.Int)/2 
// 			+ getScore(Ability.Wis)/3;
	}

	/**
	*  Returns a string representation of this persona.
	*/
	public String toString() {
		return (name == null ? "" : name + ", ")
			+ (type == null ? "" : type + " ")
			+ classString() + " "
			+ "(AC " + getAC() + ", MV " + getMV() + ", HD " + getHD() 
			+ ", hp " + getHP() + ", Atk " + attack + "; " 
			+ abilityString() + "; XP " + totalXP() + ")";
	}	

	/**
	*  String representation of all class & levels.
	*/
	public String classString() {
		String s = "";
		for (ClassRecord record: classRecordList) {
			s += record.getInfo().getAbbreviation() + record.getLevel();
			if (record != classRecordList.get(classRecordList.size()-1))
				s += "/";
		}
		return s;	
	}

	/**
	*  String representation of all ability scores.
	*/
	public String abilityString() {
		String s;
		s = "Str " + abilityScore[0] + ", Int " + abilityScore[1]
			+ ", Wis " + abilityScore[2] + ", Dex " + abilityScore[3]
			+ ", Con " + abilityScore[4] + ", Cha " + abilityScore[5];
		return s;
	}

	/**
	*  Main test method.
	*/
	public static void main (String[] args) {
		Persona p = new Persona(null, "Human", Alignment.Neutral, 
			ClassList.getInfo("Fighter"), 1);
		System.out.println(p);
	}
	
	/*--------------------------------------------------------------------------
	* TODO:
	* Properly equipped armor, shield, weapon.
	* Later: Languages, special abilities, equipment list, encumbrance.
	* Note: Not called "character" to avoid conflict with java.lang.Character.
	* Using Ability enum array saves maintenance of 12 individual methods.
	*-------------------------------------------------------------------------*/
}

